@@ -8,13 +8,18 @@ from torch.autograd import Function
 # try to import cuda
 
 try:
-    from flash_cosine_sim_attention_cuda import (
+    from flash_cosine_sim_attention_cuda_0_0_18 import (
         forward,
         backward
     )
 except ImportError:
     print('CUDA extension for flash-cosine-sim-attention was not compiled correctly - please run `pip install flash-cosine-sim-attention --force-reinstall --no-cache-dir`')
 
+# constants
+
+ALLOWED_QK_DIMS = (64,)
+ALLOWED_V_DIMS = (64,)
+
 # helper functions
 
 def exists(val):
@@ -53,7 +58,9 @@ def plain_cosine_sim_attention(
     assert not (causal and exists(mask)), 'mask should not be supplied if causality is needed'
 
     if l2norm_qk:
+        dtype = q.dtype
         q, k = map(l2norm, (q, k))
+        q, k = map(lambda t: t.type(dtype), (q, k))
 
     sim = einsum('... i d, ... j d -> ... i j', q, k)
     sim = sim * scale
@@ -87,9 +94,12 @@ class FlashCosineSimAttention(Function):
     ):
         qk_dim, v_dim = q.shape[-1], v.shape[-1]
 
+        assert (qk_dim in ALLOWED_QK_DIMS), f'query key dimension must be one of {ALLOWED_QK_DIMS}'
+        assert (v_dim in ALLOWED_V_DIMS), f'value dimension must be one of {ALLOWED_V_DIMS}'
+
         batch, heads, seq, _, dim, device, dtype = *q.shape, v.shape[-1], q.device, q.dtype
 
-        mask = default(mask, lambda: torch.ones(q.shape[0], 0, device = q.device, dtype = torch.bool))
+        mask = default(mask, lambda: torch.empty(q.shape[0], 0, device = q.device, dtype = torch.bool))
 
         attn_bias = default(attn_bias, torch.empty(1, 0, 0, device = q.device, dtype = dtype))
 
@@ -161,7 +171,9 @@ def flash_cosine_sim_attention(
     assert not (causal and exists(mask)), 'mask should not be supplied if causality is needed'
 
     if l2norm_qk:
+        dtype = q.dtype
         q, k = map(l2norm, (q, k))
+        q, k = map(lambda t: t.type(dtype), (q, k))
 
     o = FlashCosineSimAttention.apply(
         q, k, v,
